/*
 *  sx_netdev_addons - code that makes sx_netdev more useful on a general purpose
 *  Linux distribution.
 *
 *  Copyright 2016, Cumulus Networks, Inc
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms and conditions of the GNU General Public
 *  License, version 2, as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 */

#ifndef __SX_NETDEV_ADDONS_H__
#define __SX_NETDEV_ADDONS_H__

#include <net/netfilter/nf_log.h>
#include <linux/proc_fs.h>
#include <linux/port.h>
#include "sx_netdev.h"

#define NFULNL_QTHRESH_DEFAULT  100

static int sx_netdev_nflog_ing_group;
static int sx_netdev_nflog_egr_group;
static DEFINE_SPINLOCK(sx_netdev_nflog_lock);

static struct net_device *sx_netdev_get_port_netdev(u16 sysport)
{
	struct net_device *netdev;
	int i;

	for (i = 0; i < MAX_PORT_NETDEV_NUM; i++) {
		netdev = g_netdev_resources->port_netdev[PORT_TYPE_SINGLE][sysport][i];
		if (netdev)
			return netdev;
	}

	return NULL;
}

static struct net_device *sx_netdev_get_lag_netdev(u16 lag_id)
{
	struct net_device *netdev;
	int i;

	for (i = 0; i < MAX_PORT_NETDEV_NUM; i++) {
		netdev = g_netdev_resources->port_netdev[PORT_TYPE_LAG][lag_id][i];
		if (netdev)
			return netdev;
	}

	return NULL;
}

static void sx_netdev_sample_packet(struct completion_info *comp_info, void *context)
{
	struct net_device *netdev;
	struct net_device *dest_netdev;
	struct nf_loginfo li;
	u16 lag_id, sysport;

	/*
	 * HW currently supports only ingress sampling
	 */
	memset(&li, 0, sizeof(li));
	li.type              = NF_LOG_TYPE_ULOG;
	li.u.ulog.copy_len   = 0xffff;
	li.u.ulog.qthreshold = NFULNL_QTHRESH_DEFAULT;
	li.u.ulog.group      = sx_netdev_nflog_ing_group;

	if (comp_info->is_lag) {
		lag_id = comp_info->sysport;
		netdev = sx_netdev_get_lag_netdev(lag_id);
		if (!netdev) {
			sx_core_get_local(g_sx_dev, lag_id, comp_info->lag_subport, &sysport);
			netdev = sx_netdev_get_port_netdev(sysport);
		}
	} else {
		netdev = sx_netdev_get_port_netdev(comp_info->sysport);
	}

	if (comp_info->dest_is_lag) {
		lag_id = comp_info->dest_sysport;
		dest_netdev = sx_netdev_get_lag_netdev(lag_id);
		if (!dest_netdev) {
			sx_core_get_local(g_sx_dev, lag_id, comp_info->dest_lag_subport, &sysport);
			dest_netdev = sx_netdev_get_port_netdev(sysport);
		}
	} else {
		dest_netdev = sx_netdev_get_port_netdev(comp_info->dest_sysport);
	}

	if (netdev && (netdev->flags & IFF_UP)) {
		sx_netdev_fix_sx_skb(comp_info, comp_info->skb, netdev);
		nf_log_packet_queue(dev_net(netdev), NFPROTO_UNSPEC, 0,
				    comp_info->skb, netdev, dest_netdev,
				     &li, "sflow");
	}
}

/* Device nflog_group Control Proc Read Entry */
static int sx_netdev_proc_nflog_group_show(struct seq_file *m, void *v)
{
	unsigned long flags;

	spin_lock_irqsave(&sx_netdev_nflog_lock, flags);
	seq_printf(m, "Sample NFLOG group:\n");
	seq_printf(m, "  %s=%d\n", "rx", sx_netdev_nflog_ing_group);
	seq_printf(m, "  %s=%d\n", "tx", sx_netdev_nflog_egr_group);
	spin_unlock_irqrestore(&sx_netdev_nflog_lock, flags);

	return 0;
}

static int sx_netdev_proc_nflog_group_open(struct inode *inode, struct file *file)
{
	return single_open(file, sx_netdev_proc_nflog_group_show, NULL);
}

static ssize_t sx_netdev_proc_nflog_group_write(struct file *file, const char *buf,
					  size_t count, loff_t *loff)
{
	char nflog_group_str[IFNAMSIZ + 6];
	char *ptr;
	int len;
	long nflog_group;
	unsigned long flags;

	if (count >= sizeof(nflog_group_str))
		count = sizeof(nflog_group_str) - 1;

	if (copy_from_user(nflog_group_str, buf, count))
		return -EFAULT;

	nflog_group_str[count] = 0;
	ptr = strchr(nflog_group_str, '=');

	if (!ptr) {
		ptr = strchr(nflog_group_str, ':');
		if (!ptr) {
			printk(KERN_ERR PFX "sample_group syntax not recognized\n");
			return count;
		}
	}
	*ptr++ = 0;
	len = strlen(nflog_group_str);

	if (kstrtol(ptr, 0, &nflog_group))
		return -EINVAL;

	spin_lock_irqsave(&sx_netdev_nflog_lock, flags);
	if (!strncmp("rx", nflog_group_str, len))
		sx_netdev_nflog_ing_group = nflog_group;
	else if (!strncmp("tx", nflog_group_str, len))
		sx_netdev_nflog_egr_group = nflog_group;
	else
		printk(KERN_ERR PFX "Unable to set sample_group\n");
	spin_unlock_irqrestore(&sx_netdev_nflog_lock, flags);

	return count;
}

static const struct file_operations sx_netdev_proc_nflog_group_ops = {
	.owner = THIS_MODULE,
	.open = sx_netdev_proc_nflog_group_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = sx_netdev_proc_nflog_group_write,
	.release = single_release,
};

static struct proc_dir_entry *sx_netdev_proc_root;

#define SX_NETDEV_PROC_ROOT             "mlx"
#define SX_NETDEV_PROC_NFLOG_START      SX_NETDEV_PROC_ROOT "/sx_netdev"
#define SX_NETDEV_PROC_NFLOG_GROUPS     "sample_groups"

static void __init sx_netdev_proc_init(void)
{
	proc_mkdir(SX_NETDEV_PROC_ROOT, NULL);
	sx_netdev_proc_root = proc_mkdir(SX_NETDEV_PROC_NFLOG_START, NULL);

	proc_create_data(SX_NETDEV_PROC_NFLOG_GROUPS, 0600, sx_netdev_proc_root,
		&sx_netdev_proc_nflog_group_ops, NULL);

}

static void __exit sx_netdev_proc_cleanup(void)
{
	if (sx_netdev_proc_root) {
		remove_proc_entry(SX_NETDEV_PROC_NFLOG_GROUPS, sx_netdev_proc_root);
	}
	remove_proc_entry(SX_NETDEV_PROC_NFLOG_START, NULL);
	remove_proc_entry(SX_NETDEV_PROC_ROOT, NULL);
}

static void sx_get_drvinfo(struct net_device *bond_dev,
		struct ethtool_drvinfo *drvinfo)
{
	strlcpy(drvinfo->driver, "sx_netdev", sizeof(drvinfo->driver));
	strlcpy(drvinfo->version, DRV_VERSION, sizeof(drvinfo->version));
}

int sx_get_ts_info(struct net_device *dev, struct ethtool_ts_info *info);

static const struct ethtool_ops sx_cl_ethtool_ops = {
	.get_link_ksettings = port_get_link_settings,
	.set_link_ksettings = port_set_link_settings,
	.get_link = ethtool_op_get_link,
	.get_strings = port_get_strings,
	.get_ethtool_stats = port_get_ethtool_stats,
	.get_ethtool_stats_clear = port_get_ethtool_stats_clear,
	.get_sset_count = port_get_sset_count,
	.set_phys_id = port_set_phys_id,
	.get_drvinfo = sx_get_drvinfo,
	.get_module_info = port_get_module_info,
	.get_module_eeprom = port_get_module_eeprom,
	.get_fecparam = port_get_fecparam,
	.set_fecparam = port_set_fecparam,
	.get_ts_info = sx_get_ts_info,
};

static int sx_change_proto_down(struct net_device *dev, bool proto_down)
{
        dev->proto_down = proto_down;
        return 0;
}

struct rtnl_link_stats64 *sx_get_stats64(struct net_device *dev,
		    struct rtnl_link_stats64 *stats)
{
	struct sx_net_priv *priv = netdev_priv(dev);
	struct rtnl_link_stats64 sx_stats;
	struct ethtool_stats estats = { 0, };
	unsigned int start;
	int count;
	u64 *hw_stats;

	do {
		start = u64_stats_fetch_begin(&priv->syncp);
		memcpy(&sx_stats, &priv->stats64,
		       sizeof(struct rtnl_link_stats64));
	} while (u64_stats_fetch_retry(&priv->syncp, start));

	count = port_get_sset_count(dev, ETH_SS_STATS);
	if (count <= 0)
		goto err_out;

	hw_stats = kcalloc(count, sizeof(u64), GFP_ATOMIC);
	if (!hw_stats)
		goto err_out;

	estats.n_stats = count;
	port_get_ethtool_stats(dev, &estats, hw_stats);

	stats->rx_packets = hw_stats[1] + hw_stats[2] + hw_stats[3];
	stats->rx_bytes = hw_stats[0];
	stats->multicast = hw_stats[3];
	stats->rx_errors = hw_stats[12] + hw_stats[13] + hw_stats[25] + sx_stats.rx_errors;
	stats->rx_dropped = hw_stats[8] + hw_stats[9] + hw_stats[10] + hw_stats[11] + sx_stats.rx_dropped;
	stats->rx_frame_errors = sx_stats.rx_frame_errors;

	stats->tx_packets = hw_stats[5] + hw_stats[6] + hw_stats[7];
	stats->tx_bytes = hw_stats[4];
	stats->tx_errors = hw_stats[18] + sx_stats.tx_fifo_errors;
	stats->tx_dropped = hw_stats[17] + hw_stats[19] + hw_stats[20] + sx_stats.tx_dropped;

	kfree(hw_stats);
err_out:
	return stats;
}

#endif /* __SX_NETDEV_ADDONS_H__ */
